"""Compute 2pt stats in zeldovich approximation

"""
#imports
from scipy.interpolate import InterpolatedUnivariateSpline as interpolate
from scipy.misc import derivative
import pyfftw
from scipy.special import loggamma
from scipy.interpolate import interp1d
import numpy as np

def loginterp(x, y, yint = None, side = "both", lorder = 9, rorder = 9, lp = 1, rp = -1,
              ldx = 1e-6, rdx = 1e-6):
    '''
    Extrapolate function by evaluating a log-index of left & right side.
    Use 9th order FD derivative to find effective power law index and use that power law for extrapolation.
    From Chirag Modi's CLEFT code at
    https://github.com/modichirag/CLEFT/blob/master/qfuncpool.py
    The warning for divergent power laws on both ends is turned off.

    Parameters:
    -----------
    x: array
        abscissa values
    y: array
        funtion values
    yint: callable,optional
        interpolator callable
    side: string, optional:
        one of 'l','r',or 'both' - edge to extrapolate
    lorder,rorder: int, optional
        left and right order of FD derivative evalauted at edges
    lp,rp: int,optional
        left and right extrap indices, assumes first and last elements
    lidx,ridx: float,optional
        left and right FD epsilon value

    Returns:
    -----------
    callable
        interpolated univariate spline over the extrapolated range
    '''
    if yint is None:
        yint = interpolate(x, y, k = 5)
    if side == "both":
        side = "lr"

    #get derivatives
    l =lp
    r =rp
    lneff = derivative(yint, x[l], dx = x[l]*ldx, order = lorder)*x[l]/y[l]
    rneff = derivative(yint, x[r], dx = x[r]*rdx, order = rorder)*x[r]/y[r]

    #the exrapolated power laws
    xl = np.logspace(-12, np.log10(x[l]), 10**5)
    xr = np.logspace(np.log10(x[r]), 12., 10**5)
    yl = y[l]*(xl/x[l])**lneff
    yr = y[r]*(xr/x[r])**rneff

    #put it together
    xint = x[l+1:r].copy()
    yint = y[l+1:r].copy()
    if side.find("l") > -1:
        xint = np.concatenate((xl, xint))
        yint = np.concatenate((yl, yint))
    if side.find("r") > -1:
        xint = np.concatenate((xint, xr))
        yint = np.concatenate((yint, yr))
    #IUS
    yint2 = interpolate(xint, yint, k = 5, ext=3)

    return yint2

class SphericalBesselTransform:
    '''
    Spherical bessel function class with FFTW option added. Based on Stephen Chen's velocileptor's (https://github.com/sfschen/velocileptors) version of Chirag Modi's CLEFT code (https://github.com/modichirag/CLEFT).

    Class to perform spherical bessel transforms via FFTLog for a given set of qs, ie.
    the untransformed coordinate, up to a given order L in bessel functions (j_l for l
    less than or equal to L. The point is to save time by evaluating the Mellin transforms
    u_m in advance.

    Uses pyfftw, which can perform multiple (ncol) Fourier transforms at once, one for
    each bias contribution.

    Based on Yin Li's package mcfit (https://github.com/eelregit/mcfit)
    with the above modifications.

    '''

    def __init__(self, qs, L=15, ncol = 1, low_ring=True, fourier=False, threads=1,
                 import_wisdom=False, wisdom_file='./fftw_wisdom.npy',useFFTW=True):
        '''
        Parameters:
        ---------
        qs: array,float
            Wavenumber abscissa pts
        L: int,optional
            Maximum order in \ell used for the angular sum used to compute P_ZA as 1d integrals (Schneider & Bartelmann 1995)
        ncol: int,optional
            Used to compute multiple transforms at once for different biasing terms, for za will be 1
        low_ring: boolean, optional
            Whether to apply lowring condition (force continuity when folding period, https://jila.colorado.edu/~ajsh/FFTLog/#lowring)
        fourier: boolean,optional
            If true assume pi factor for an integral in fourier space
        import_wisdom: boolean, optional
            Whether to use saved setup for the fftw to reduce overhead
        wisdom_file: string, optional
            The filename for the above
        useFFTW: boolean, optional
            Whether to use pyfftw (default) or numpy ffts
        '''
        # numerical factor of sqrt(pi) in the Mellin transform
        # if doing integral in fourier space get in addition a factor of 2 pi / (2pi)^3
        if not fourier:
            self.sqrtpi = np.sqrt(np.pi)
        else:
            self.sqrtpi = np.sqrt(np.pi) / (2*np.pi**2)

        self.q = qs
        self.L = L
        self.ncol = ncol

        self.Nx = len(qs)
        self.Delta = np.log(qs[-1]/qs[0])/(self.Nx-1)
        self.useFFTW = useFFTW

        # zero pad the arrays to the preferred length format for ffts, 2^N
        self.N = 2**(int(np.ceil(np.log2(self.Nx))) + 1)
        self.Npad = self.N - self.Nx
        if(self.useFFTW): #fftw setup
            self.ii_l = self.Npad - self.Npad//2 # left and right indices sandwiching the padding
            self.ii_r = self.N - self.Npad//2
                    # Set up FFTW objects:
            if import_wisdom:
                pyfftw.import_wisdom(tuple(np.load(wisdom_file)))

            self.fks = pyfftw.empty_aligned((self.ncol,self.N//2 + 1), dtype='complex128')
            self.fs  = pyfftw.empty_aligned((self.ncol,self.N), dtype='float64')

            self.gks = pyfftw.empty_aligned((self.ncol,self.N//2 + 1), dtype='complex128')
            self.gs  = pyfftw.empty_aligned((self.ncol,self.N), dtype='float64')

            pyfftw.config.NUM_THREADS = threads
            self.fft_object = pyfftw.FFTW(self.fs, self.fks, direction='FFTW_FORWARD',threads=threads)
            self.ifft_object = pyfftw.FFTW(self.gks, self.gs, direction='FFTW_BACKWARD',threads=threads)
            # Set up FFTW objects:
            if import_wisdom:
                pyfftw.import_wisdom(tuple(np.load(wisdom_file)))
        else:
            self.pads = np.zeros( (self.N-self.Nx)//2  )
            self.pad_iis = np.arange(self.Npad - self.Npad//2, self.N - self.Npad//2)

        # Set up the FFTLog kernels u_m up to, but not including, L
        ms = np.arange(0, self.N//2+1)
        self.ydict = {}; self.udict = {}; self.qdict= {}

        #L is the number of orders to track through for bessel functions
        #Recall that lowringing is setting the bdr values of lnxy to be equal so that the period folds continuously
        #Here this is just done L times because we will need that many bessel kernels in the angular integration
        if low_ring:
            for ll in range(L):
                q = max(0, 1.5 - ll)
                lnxy = self.Delta/np.pi * np.angle(self.UK(ll,q+1j*np.pi/self.Delta)) #ln(xmin*ymax)
                ys = np.exp( lnxy - self.Delta) * qs/ (qs[0]*qs[-1])
                us = self.UK(ll, q + 2j * np.pi / self.N / self.Delta * ms) \
                        * np.exp(-2j * np.pi * lnxy / self.N / self.Delta * ms)
                if(self.useFFTW): us[self.N//2] = us[self.N//2].real # manually impose low ring

                self.ydict[ll] = ys; self.udict[ll] = us; self.qdict[ll] = q

        else:
            # if not low ring then just set x_min * y_max = 1
            for ll in range(L):
                q = max(0, 1.5 - ll)
                ys = np.exp(-self.Delta) * qs / (qs[0]*qs[-1])
                us = self.UK(ll, q + 2j * np.pi / self.N / self.Delta * ms)
                if(self.useFFTW): us[self.N//2] = us[self.N//2].real # manually impose low ring

                self.ydict[ll] = ys; self.udict[ll] = us; self.qdict[ll] = q


    def export_wisdom(self, wisdom_file='./fftw_wisdom.npy'):
        '''Parameters:
         ----------
         wisdom_file: string,optional
            Wisdom filename
         '''
        np.save(wisdom_file, pyfftw.export_wisdom())

    def sph(self, nu, fq):
        '''
        The workhorse of the class. Spherical Hankel Transforms fq on coordinates self.q.
        Parameters:
        ----------
        nu: int
            Order of spherical bessel function
        fq: array (float)
            Values of function to transform
        Returns:
        ----------
        (array (float), array (float))
            Tuple of the (absissca, function) values in the transformed space
        '''
        q = self.qdict[nu]; y = self.ydict[nu] #pre-tabulated for nus we want
        useFFTW = self.useFFTW
        if(useFFTW):
            self.fs[:] = 0 # on NERSC this seems necessary or this variable spills over from previous calls
            self.fs[:,self.Npad - self.Npad//2 : self.N - self.Npad//2] = fq * self.q**(3-q)
            fks = self.fft_object()
            self.gks[:] = np.conj(fks * self.udict[nu])
            gs = self.ifft_object()
            return y, gs[:,self.ii_l:self.ii_r] * y**(-q)
        else:
            f = np.concatenate( (self.pads,self.q**(3-q)*fq,self.pads) )
            fks = np.fft.rfft(f)
            gks = self.udict[nu] * fks
            gs = np.fft.hfft(gks) / self.N
            return y, y**(-q) * gs[self.pad_iis]

    def UK(self, nu, z):
        '''
        The Mellin transform of the spherical bessel transform. - from mcfit
        Parameters:
        ----------
        nu: int
            Order of spherical bessel function
        z: float
            Argument
        Returns:
        ----------
        array (float)
            Spherical bessel kernel evaluated at z
        '''
        return self.sqrtpi * np.exp(np.log(2)*(z-2) + loggamma(0.5*(nu+z)) - loggamma(0.5*(3+nu-z)))

#Qfunc for computing the power - tossing almost everything from CLEFT
class QFuncFFT:
    '''
       Chirag's notes:
       Class to calculate all the functions of q, X(q), Y(q), U(q), xi(q) etc.
       as well as the one-loop terms Q_n(k), R_n(k) using FFTLog.
       Throughout we use the ``generalized correlation function'' notation of 1603.04405.
       Note that one should always cut off the input power spectrum above some scale.
       I use exp(- (k/20)^2 ) but a cutoff at scales twice smaller works equivalently,
       and probably beyond that. The important thing is to keep all integrals finite.
       This is done automatically in the Zeldovich class.
       Currently using the numpy version of fft. The FFTW takes longer to start up and
       the resulting speedup is unnecessary in this case.

    '''
    def __init__(self, k, p, qv = None):
        '''This assumes L=5 (with lowring) for the angular evaluation
        Parameters:
        ----------
        k: array (float)
            Wavenumbers
        p: array (float)
            Power spectrum values
        qv: array (float),optional
            q values to use when computing the generalized correlation functions
        '''
        self.k = k
        self.p = p
        if qv is None:
            self.qv = np.logspace(-5,5,20000) #can decrease?
        else:
            self.qv = qv

        self.sph = SphericalBesselTransform(self.k, L=5, low_ring=True, fourier=True,
                                            useFFTW=False) #keeping with old choice
        self.setup_xiln()
        self.setup_2pts()

    def setup_xiln(self):
        # Compute a bunch of generalized correlation functions
        self.xi00 = self.xi_l_n(0,0)
        self.xi1m1 = self.xi_l_n(1,-1)
        self.xi0m2 = self.xi_l_n(0,-2, side='right') # since this approaches constant on the left only interpolate on right
        self.xi2m2 = self.xi_l_n(2,-2)

    def setup_2pts(self):
        # Piece together xi_l_n into what we need
        self.Xlin = 2./3 * (self.xi0m2[0] - self.xi0m2 - self.xi2m2)
        self.Ylin = 2 * self.xi2m2
        self.Ulin = - self.xi1m1
        self.corlin = self.xi00

    def xi_l_n(self, l, n, _int=None, extrap=False, qmin=1e-3, qmax=1000, side='both'):
        '''
        Calculates the generalized correlation function xi_l_n (Schmittful et al 16), which is xi when l = n = 0
        If _int is None assume integrating the power spectrum.
        Parameters:
        ----------
        l: int
            Legendre expansion order
        n: int
            Power on q in the integrand
        Returns:
        ----------
        array (float)
            Xi_l_n evaluated at qv
        '''
        if _int is None:
            integrand = self.p * self.k**n
        else:
            integrand = _int * self.k**n

        qs, xint =  self.sph.sph(l,integrand)

        if extrap:
            qrange = (qs > qmin) * (qs < qmax)
            return loginterp(qs[qrange],xint[qrange],side=side)(self.qv)
        else:
            return np.interp(self.qv, qs, xint)

#Stripped down CLEFT class
class CLEFT:
    '''
    jms - Removing everything but linear
    Class to calculate power spectra up to one loop.
    Based on Chirag's code
    https://github.com/sfschen/velocileptors/blob/master/LPT/cleft_fftw.py
    The bias parameters are ordered in pktable as 1
    '''

    def __init__(self, k, p, cutoff=10, jn=5, N = 2000, threads=1, extrap_min = -5, extrap_max = 3, import_wisdom=False, wisdom_file='wisdom.npy'):
        '''
        Parameters:
        ----------
        k: array (float)
            Wavenumber array
        p: array (float)
            Power array
        cutoff: float
            k-scale of exponential suppression to allow for well-behaved FFT
        jn: int,optional
            Maximum order in \ell used for the angular sum used to compute P_ZA as 1d integrals (Schneider & Bartelmann 1995)
        N: int,optional
            Number of points used in extrap abscissa array
        threads: int,optional
            Num threads passed to pyfftw
        extrap_min: int,optional
            Min in logk used in extrap array
        extrap_max: int,optional
            Max in logk used in extrap array
        import_wisdom: boolean, optional
            Whether to use saved setup for the fftw to reduce overhead
        wisdom_file: string, optional
            The filename for the above

        '''
        self.N = N
        self.extrap_max = extrap_max
        self.extrap_min = extrap_min

        self.cutoff = cutoff
        self.kint = np.logspace(extrap_min,extrap_max,self.N)
        self.qint = np.logspace(-extrap_max,-extrap_min,self.N)

        self.update_power_spectrum(k,p)

        self.pktable = None
        self.num_power_components = 1 #since just ZA


        self.jn = jn
        self.threads = threads
        self.import_wisdom = import_wisdom
        self.wisdom_file = wisdom_file
        self.sph = SphericalBesselTransform(self.qint, L=self.jn, ncol=self.num_power_components,
                                            threads=self.threads, import_wisdom= self.import_wisdom,
                                            wisdom_file = self.wisdom_file)
        #uses FFTW version by default

    def update_power_spectrum(self, k, p):
        # Updates the power spectrum and various q functions. Can continually compute for new cosmologies without reloading FFTW
        self.k = k
        self.p = p
        self.pint = loginterp(k,p)(self.kint) * np.exp(-(self.kint/self.cutoff)**2)
        self.setup_powerspectrum()

    def setup_powerspectrum(self):
        # This sets up terms up to one looop in the combination (symmetry factors) they appear in pk
        self.qf = QFuncFFT(self.kint, self.pint, qv=self.qint)

        # linear terms
        self.Xlin = self.qf.Xlin
        self.Ylin = self.qf.Ylin

        self.XYlin = self.Xlin + self.Ylin; self.sigma = self.XYlin[-1]
        self.yq = self.Ylin / self.qint
        self.corlin = self.qf.corlin

    def p_integrals(self, k):
        '''
        Compute P(k) for a single k as a vector of all bias contributions.
        Parameters:
        ----------
        k: array (float)
            Wavenumber array
        Returns:
        ----------
        array
            ZA power
        '''
        ksq = k**2
        expon = np.exp(-0.5*ksq * (self.XYlin - self.sigma))
        suppress = np.exp(-0.5 * ksq *self.sigma)

        ret = np.zeros(self.num_power_components)

        bias_integrands = np.zeros( (self.num_power_components,self.N)  )

        for l in range(self.jn):
            bias_integrands[-1,:] = 1 # this is the counterterm, minus a factor of k2
            #this is the ZA
            # multiply by IR exponent
            if l == 0:
                bias_integrands = bias_integrands * expon
                bias_integrands -= bias_integrands[:,-1][:,None] # note that expon(q = infinity) = 1
            else:
                bias_integrands = bias_integrands * expon * self.yq**l

            # do FFTLog
            ktemps, bias_ffts = self.sph.sph(l, bias_integrands)
            ret +=  k**l * interp1d(ktemps, bias_ffts)(k)

        return 4*suppress*np.pi*ret

    def make_ptable(self, kmin = 1e-3, kmax = 3, nk = 100):
        '''
        Make a table of different terms of P(k) between a given
        'kmin', 'kmax' and for 'nk' equally spaced values in log10 of k
        This is the most time consuming part of the code.
        Parameters:
        ----------
        kmin: float,optional
            Min wavenumber
        kmax: float,optional
            Max wavenumber
        nk: int,optional
            Number of pts
        Returns:
        ----------
        None
        '''
        self.pktable = np.zeros([nk, self.num_power_components+1]) # one column for ks
        kv = np.logspace(np.log10(kmin), np.log10(kmax), nk)
        self.pktable[:, 0] = kv[:]
        for foo in range(nk):
            self.pktable[foo, 1:] = self.p_integrals(kv[foo])


    def export_wisdom(self, wisdom_file='./wisdom.npy'):
        """
        Wrapper for the SBT class wisdom saving function
        Parameters:
        ----------
        wisdom_file: string,optional
            Filename for wisdom.
        """
        self.sph.export_wisdom(wisdom_file=wisdom_file)

    #From Stephen's gsm model class, appropriated here for xiza since it is all I want
    def compute_xi_real(self, rr, kswitch=1e-2,kint= np.logspace(-5,3,4000)):
        '''
        Compute the real-space correlation function at rr.
        '''
        #kint = np.logspace(-5,3,4000)
        plin = loginterp(self.k, self.p)(kint)
        # This is just the zeroth moment:
        kv   = self.pktable[:,0]
        pzel = self.pktable[:,-1]

        peft = pzel
        weight =  0.5 * (1 + np.tanh(3*np.log(kint/kswitch))) #smooth away low-k to replace with plin below
        weight[weight < 1e-3] = 0

        _integrand = loginterp(kv, peft)(kint) #at some point turn into theory extrapolation
        _integrand = weight * _integrand + (1-weight) * (plin)

        sph_xi = SphericalBesselTransform(kint,L=1,fourier=True,
                                            useFFTW=False) #keeping with old choice, for whatever reason
        #dropping L to 1 since we won't be needing higher here for linear, only using sph for j0 transform to xi

        qint, xi = sph_xi.sph(0,_integrand)

        xir = interpolate(qint,xi)(rr)
        return xir



'''Real-space tracer/galaxy/halo model'''
from gzpt import hzpt
from gzpt.bb import *
import numpy as np
from numpy import inf
from scipy.special import erf,gamma,poch
from scipy.interpolate import InterpolatedUnivariateSpline as ius
from scipy.integrate import cumtrapz

class AutoCorrelator(hzpt):
    def __init__(self,params,hzpt,params_exc=None,params_sat=None):
        '''
        Parameters
        ----------
        params : list (float)
            List of current hzpt parameter values.
        hzpt : hzpt
            Hzpt base class that holds the linear, zeldovich, and z information.
        params_exc : list,optional
            Exclusion parameters, if desired.
        params_sat : list,optional
            Satellite parameters, if desired.
        '''
        self.params = params
        self.nmax = min((len(self.params[2:])-1)//2,2)
        self.useExc,self.useSat = False,False
        if(params_exc is not None):
            self.useExc = True
            for p in params_exc: self.params.append(p)
        if(params_sat is not None):
            self.useSat = True
            for p in params_sat: self.params.append(p)
        self.hzpt = hzpt


    def Fk_excl(self,k,R_excl,wantGrad=False):
        '''
        Model for the exclusion step.
        Use the Exp model.
        Parameters
        ----------
        k : array (float)
            pts to evaluate P
        R_excl : float
            Exclusion radius
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        array or (array,array)
            Exclusion kernel, and gradient, if asked for
        '''

        def _hyp0f2(b1,b2, z, eps=1e-6, nmax=10):
            sum = 0
            #accumulate the sum from scratch, no convenient identities, but 5 terms seems good enough
            #mpmath does this but don't want to introduce dependency just for one function
            for k in range(nmax):
                sum+= 1/(poch(b1,k)*poch(b2,k)) * z**k / np.math.factorial(k)
            return sum

        def _fk(k,R_excl):
            #see B.8 of paper
            #try to come back and pass arguments for 0F2 convergence
            t1 = R_excl**3 / 3 * gamma(7/4) * _hyp0f2(1/2,5/4,(k*R_excl/4)**4)
            t2 = -k**2 * R_excl**5 /24 * gamma(5/4) * _hyp0f2(3/2,7/4,(k*R_excl/4)**4)
            return (t1 + t2)/(2*np.pi**2)

        F = -2*_fk(k,R_excl) + _fk(k,2**(-1/4) * R_excl)
        #with these settings can't trust k>10/R h\Mpc
        #Obviously this is not ideal - should use \xi exclusion!
        if(np.any(k>10/R_excl)): F[k>10/R_excl] = 0
        if(wantGrad):
            #analytic gradient of the above expression
            t1 = -R_excl**2 * gamma(7/4) * _hyp0f2(1/2,5/4,k*R_excl/4)
            t2 = -1/210 * k**4 * R_excl**6 * gamma(7/4) * _hyp0f2(3/2,9/4,k*R_excl/4)
            t3 = 1/6 * k**2 * R_excl**4 * gamma(9/4) * _hyp0f2(3/2,7/4,k*R_excl/4)
            t4 = 1/5040 * gamma(9/4) * k**6 * R_excl**8 * _hyp0f2(5/2,11/4,k*R_excl/4)
            grad_R = t1 + t2 + t3 + t4
            if(np.any(k>10)): grad_R[k>10] = 0
            grad_F = [grad_R]
            return F,grad_F
        else:
            return F


    def convolve(self,k,x,y):
        #FIXME most naive implementation, could do better with just about anything else for the integral
        #Do the N^2 operation, assume (probably bad) that not zero padding is fine
        #It would be better (for all N>100) to do the fft convolution
        A,B = ius(k,x,ext=1),ius(k,y,ext=1)
        #logdelta = np.log(k[1:]/k[:-1])
        N = len(k)
        res = np.zeros(N)
        for i in range(N):
            res[i] = np.trapz(A(abs(k-k[i]))*B(k),x=k)
        return res
        #qualitatively seems to do okay, but should really come back and fix it

    def Power(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            Power spectrum in k (h/Mpc)
        '''
        #Do some parameter juggling - unfortunate that this is repeated for Xi, FIXME
        if(self.useExc and self.useSat):
            nbar,b1,pparams,eparams = self.params[0],self.params[1],self.params[2:2+(2*self.nmax)+1],self.params[2+(2*self.nmax)+1:-2]
        elif(self.useExc and ~self.useSat):
            nbar,b1,pparams,eparams = self.params[0],self.params[1],self.params[2:2+(2*self.nmax)+1],self.params[2+(2*self.nmax)+1:]
        elif(~self.useExc and self.useSat):
            nbar,b1,pparams = self.params[0],self.params[1],self.params[2:2+(2*self.nmax)+1]
        else:
            nbar,b1,pparams = self.params[0],self.params[1],self.params[2:]

        def Pk(k):
            #Set BB
            if(wantGrad):
                grads=[]
                bb,bbgrad = PBB(k,pparams,nmax=self.nmax,wantGrad=True)
                p = 1/nbar + b1**2 * (self.hzpt.P_zel(k) + bb)
                nbar_grad = -(1/nbar**2) *np.ones(len(k))
                b1_grad = 2*(p- 1/nbar)/b1
                grads.append(np.atleast_2d(nbar_grad).T)
                grads.append(np.atleast_2d(b1_grad).T)
                grads.append(b1**2 * bbgrad)
            else:
                bb = PBB(k,pparams,nmax=self.nmax)
                p = 1/nbar + b1**2 * (self.hzpt.P_zel(k) + bb)

            #Set exclusion
            if(self.useExc):
                if(len(eparams)>1):
                    raise NotImplementedError("Power exclusion is only implemented for the Exp model. Please use FFTLog.")
                R=eparams[0]
                pc = p-1/nbar
                if(wantGrad):
                    exclusion_func,e_f_grad = self.Fk_excl(k,R,wantGrad=True)
                    grad_Re = e_f_grad + self.convolve(k,e_f_grad,pc)
                    grads.append(np.atleast_2d(grad_Re).T)
                else:
                    exclusion_func =  self.Fk_excl(k,R,wantGrad=False)
                exclusion = exclusion_func + self.convolve(k,exclusion_func,pc)
            else:
                exclusion =0  #no exclusion (additive)

            #put it together
            if(wantGrad): return p + exclusion, np.hstack(grads)
            else: return p + exclusion
        return Pk


    def Fr_excl(self,r,R_excl,sigma_excl=None,wantGrad=False):
        '''
        Model for the exclusion step.
        If no sigma_excl is provided, use the Exp model, o.w. use model of Baldauf++2013.
        Parameters
        ----------
        r : array (float)
            pts to evaluate xi
        R_excl : float
            Exclusion radius
        sigma_excl:
            Exclusion step width for error function
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        array or (array,array)
            Exclusion kernel, and gradient, if asked for
        '''
        #Exp model from Appendix B.1
        if(sigma_excl==None):
            f = 1-np.exp(-(r/R_excl)**4)
            F = f**2
            if(wantGrad):
                grad_Re = -8*(r / R_excl)**4 /R_excl * np.exp(-(r/R_excl)**4) * f
                grad_F = [grad_Re,None]
                return F,grad_F
            else:
                return F
        #ExpLog
        #Error function for the exclusion step as in Baldauf++ 2013 eqns (C2), (C4), with log10 as in main text
        else:
            F = .5*(1 + erf(np.log10(r/R_excl)/(np.sqrt(2)*sigma_excl)))
            if(wantGrad):
                ln10 = np.log(10)
                exp = -np.exp(-np.log(r/R_excl)**2 / (2*sigma_excl**2 * ln10**2))
                denom = np.sqrt(2.*np.pi)* sigma_excl * ln10
                grad_Re = exp*denom/R_excl
                grad_sigmae = exp*denom*np.log(r/R_excl)/sigma_excl
                grad_F = [grad_Re,grad_sigmae]
                return F,grad_F
            else:
                return F

    def Xi(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            2pcf function of r (Mpc/h)
        '''
        if(self.useExc and self.useSat):
            b1,pparams,eparams,sparams = self.params[1],self.params[2:2+(2*self.nmax)+1],self.params[2+(2*self.nmax)+1:-2],self.params[-2:]
        elif(self.useExc and ~self.useSat):
            b1,pparams,eparams = self.params[1],self.params[2:2+(2*self.nmax)+1],self.params[2+(2*self.nmax)+1:]
            sparams=None
        elif(~self.useExc and self.useSat):
            b1,pparams,sparams = self.params[1],self.params[2:2+(2*self.nmax)+1],self.params[2+(2*self.nmax)+1:]
            eparams=None
        else:
            b1,pparams,eparams = self.params[1],self.params[2:],None
            eparams,sparams=None,None

        def xi(r):
            #Set BB
            if(wantGrad):
                grads=[]
                bb,bbgrad = XiBB(r,pparams,nmax=self.nmax,wantGrad=True)
                xic = b1**2 * (self.hzpt.Xi_zel(r) + bb)
                b1_grad = 2.*xic/b1
                grads.append(np.atleast_2d(b1_grad).T)
                grads.append(b1**2 * bbgrad)
            else:
                bb = XiBB(r,pparams,nmax=self.nmax)
                xic = b1**2 * (self.hzpt.Xi_zel(r) + bb)


            #Set exclusion
            if(self.useExc):
                if(len(eparams)==2): #check for Exp or ExpLog model
                    R,sigma = eparams
                else:
                    R,sigma = eparams,None
                if(wantGrad):
                    exclusion,e_grad = self.Fr_excl(r,R,sigma,wantGrad=True)
                    grad_Re,grad_sigmae = e_grad
                    grads.append(np.atleast_2d(grad_Re*(1. + xic)).T)
                    grads.append(np.atleast_2d(grad_sigmae*(1. + xic)).T)
                else:
                    print('R,sigma', R,sigma)
                    exclusion = self.Fr_excl(r,R,sigma)
            else:
                exclusion=1 #no exclusion (multiplicative)

            #Set sattelites
            if(self.useSat):
                R_s_fixed=1e3 #fix compensation to effectively infinity
                A1s,R1h1s = sparams
                if(wantGrad):
                    satellite, satellite_grad = XiBB(r,[A1s,R_s_fixed,R1h1s],nmax=1,wantGrad=True)
                    grads.append(b1**2 * np.array([satellite_grad[:,0],satellite_grad[:,1]]).T)
                else:
                    satellite = XiBB(r,[A1s,R_s_fixed,R1h1s],nmax=1)
            else:
                satellite=0 #no sats (additive)

            #put it together
            xi_d = exclusion*(1. + xic) - 1.
            if(wantGrad): return xi_d + satellite, np.hstack(grads)
            else: return xi_d + satellite
        return xi

    def wp(self,rp,pi=np.linspace(0,100,100+1),rMin=0.01,wantGrad=False):
        """Projected correlation function.
        Parameters:
        -----------
        rp: array (float)
            2D abscissa values for wp
        pi_bins: array (float)
            Array must be of size that divides evenly into r - projection window, tophat for now
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns:
        ----------
        (array (float), array (float), [array (float)])
            projected radius rp, projected correlation function wp, gradient if wanted
        """

        if(wantGrad):
            wp_grad=np.zeros((len(wp),len(self.params)))
            xir,grad_xir = self.Xi(wantGrad=wantGrad)
        else:
            xir = self.Xi(wantGrad=wantGrad)
        dpi = pi[1]-pi[0] #bin width
        wp = np.zeros(len(rp))
        for i in range(len(wp)):
            rev=np.sqrt(rp[i]**2 + pi**2)
            wp[i]= 2*ius(rev,xir(rev)/np.sqrt(1- (rp[i]/(rev+rMin))**2)).integral(rev.min(),rev.max())
            if(wantGrad): wp_grad[i] =2*ius(rev,grad_xir(rev)/np.sqrt(1- (rp[i]/(rev+rMin))**2)).integral(rev.min(),rev.max())

        if(wantGrad):
            return wp,wp_grad
        else:
            return wp



class CrossCorrelator(hzpt):

    def __init__(self,params,hzpt):
        #Set maximum Pade expansion order
        self.params = params
        self.nmax = (len(self.params[1:])-1)//2
        self.hzpt = hzpt

    def Power(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            Power spectrum in k (h/Mpc)
        '''
        def Pk(k):
            b1,pparams = self.params[0],self.params[1:]
            if(wantGrad):
                bb,bbgrad = PBB(k,pparams,nmax=self.nmax,wantGrad=True)
                p = b1*(self.hzpt.P_zel(k) + bb)
                b1_grad = p/b1
                return p, np.hstack([np.atleast_2d(b1_grad).T,b1*bbgrad])
            else: return b1*(self.hzpt.P_zel(k) + PBB(k,pparams,nmax=self.nmax))
        return Pk

    def Xi(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            2pcf function of r (Mpc/h)
        '''
        def xi(r):
            b1,pparams = self.params[0],self.params[1:]
            if(wantGrad):
                bb,bbgrad = XiBB(r,pparams,nmax=self.nmax,wantGrad=True)
                xi = b1*(self.hzpt.Xi_zel(r) + bb)
                b1_grad = xi/b1
                return xi, np.hstack([np.atleast_2d(b1_grad).T, b1*bbgrad])
            else: return b1*(self.hzpt.Xi_zel(r) + XiBB(r,pparams,nmax=self.nmax))
        return xi


    '''Copying for now - will move projected statistics to another file later'''
    def wp(self,rp,pi=np.linspace(0,100,100+1),rMin=0.1,wantGrad=False):
        "FIXME: Lazy copy from AutoCorrelator - should make this function accessible by both. - general correlator class..."
        """Projected correlation function.
        Parameters:
        -----------
        rp: array (float)
            2D abscissa values for wp
        pi_bins: array (float)
            Array must be of size that divides evenly into r - projection window, tophat for now
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns:
        ----------
        (array (float), array (float), [array (float)])
            projected radius rp, projected correlation function wp, gradient if wanted
        """

        if(wantGrad):
            wp_grad=np.zeros((len(rp),len(self.params)))
            xir,grad_xir = self.Xi(wantGrad=wantGrad)
        else:
            xir = self.Xi(wantGrad=wantGrad)

        dpi = pi[1]-pi[0] #bin width
        wp = np.zeros(len(rp))
        for i in range(len(wp)):
            rev=np.sqrt(rp[i]**2 + pi**2)
            wp[i]= 2*ius(rev,xir(rev)/np.sqrt(1- (rp[i]/(rev+rMin))**2)).integral(rev.min(),rev.max())
            if(wantGrad): wp_grad[i] =2*ius(rev,grad_xir(rev)/np.sqrt(1- (rp[i]/(rev+rMin))**2)).integral(rev.min(),rev.max())


        if(wantGrad):
            return wp,wp_grad
        else:
            return wp

    def Delta_Sigma(self,rp,DS0, #PM
                    Ds,Dl,zl,rhom, #background/bin-dependent quantities
                    pi=np.linspace(0,100,100+1),rpMin=1.,num_rpint=1000,wantGrad=False):

        """Delta Sigma GGL statistic. Lens redshift is assumed to be the CrossCorrelator attribute z.
        Simple top-hat redshift distribution.
        Parameters:
        -----------
        rp: array (float)
            2D abscissa values for wp/DS
        DS0: float
            Point mass marginalization parameter for Delta Sigma (see Singh ++ 2018)
        Ds: float
            Comoving distance (flat) to source
        Dl: float
            Comoving distance (flat) to lens
        zl: float
            Lens redshift
        rhom: float
            Mean matter density at z=0, in Msun^1 pc^-2 h^1 (pc, NOT Mpc)
        pi_bins: array (float),optional
            Array must be of size that divides evenly into r - projection window, tophat for now
        rpMin: float,optional
            Minimum scale to which model is trusted, below this activate point mass.
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns:
        ----------
        (array (float), array (float), [array (float)])
            projected radius rp, projected correlation function wp, gradient if wanted
        """
        #FIXME: do not actually need this (it works though)- we are not computing tangential shear...might later
        # def inv_Sigma_crit(Ds,Dl,zl):
        #     c=2.99792e5 #km/s
        #     G=4.30071e-9 #Mpc (km/s)**2 * M_sun^-1
        #     #Assume Ds>Dl
        #     #can come back to computing Ds, DL internally  after caching distances in gzpt object, for now require input
        #     if(Ds>Dl):
        #         Dfactor = Dl*(Ds-Dl)/Ds #Mpc/h
        #     else:
        #         Dfactor = 0.
        #     pre = (1+zl)*(4*np.pi*G)/c**2 #Mpc *Msun^-1
        #     return pre*Dfactor # #Mpc^2 Msun^-1 h^-1, what we need to match units of Sigma

        #FIXME: internally check units on this, works in the example though

        if(wantGrad):
            wpgm,grad_wpgm = self.wp(rp,pi=pi,wantGrad=True)

        else:
            wpgm = self.wp(rp,pi=pi)
        #rp interpolation pts
        rpint=np.logspace(np.log10(rp.min()),np.log10(rp.max()),num_rpint) #FIXME probably not necessary to use 1000

        if(wantGrad):
            I_grad_hzpt = np.zeros((len(rpint),len(self.params)))
            I_grad_PM = np.zeros(len(rpint))

        wint = np.interp(rpint,rp,wpgm)
        rpint_c = rpint[rpint>=rpMin]
        wint_c = wint[rpint>=rpMin]
        s_int = rhom*np.interp(rpint,rp,wpgm)
        s = rhom*wpgm
        sint_c = rhom*wint_c
        t1=np.zeros(len(rpint))
        t1[rpint<rpMin] = 0.
        #Using Singh 2018 eqn 29 - there is a typo so add missing factor of 2
        t1[rpint>=rpMin] = (2./rpint_c**2)*cumtrapz(rpint_c*sint_c,x=rpint_c,initial=0)
        t2 = -s_int
        S0=np.interp(rpMin,rpint,s_int)
        t3 = (rpMin/rpint)**2 * (DS0 + S0) #compute Sigma_0 from Singh 18 eqn. 30
        DS=t1+t2+t3

        if(wantGrad): #FIXME: test for this
            grad_wpgm_int_c = np.zeros((len(rpint_c),grad_wpgm.shape[1]))
            for i in range(grad_wpgm.shape[1]):
                grad_wpgm_int[:,i] = np.interp(rpint,rp,grad_wpgm[:,i])
                grad_wpgm_int_c[:,i] = grad_wpgm_int[:,i][rpint>=rpMin]
                term1_grad[:,i] = (2./rpint**2)*cumtrapz(rpint_c*rhom*grad_wpgm_int_c,x=rpint_c,axis=0,initial=0)
            term2_grad = -rhom*grad_wpgm
            I_grad_hzpt = term1_grad+term2_grad
            I_grad_PM = 1/rpint**2

        if(wantGrad):
            grad_DS = np.concatenate([I_grad_hzpt,np.atleast_2d(I_grad_PM).T],axis=1)
            return DS,grad_DS
        else:
            return DS



#Misc. functions
from nbodykit.cosmology import Planck15
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as ius

#true utils
def W_TH(k,R):
    '''Top hat window
    Input: wavenumber k - arraylike
    position r - arraylike
    option (k,r) - for Fourier or configuration space top hat
    R - the filtering scale
    '''
    x=k*R
    return (3/x**3)*(np.sin(x) - x*np.cos(x))


def W_TH_real(r,R):
    '''Top hat window
    Input:
    position r - arraylike
    R - the filtering scale
    '''
    V = 4*np.pi/3 *R**3
    if(len(r)>0):
        indicator = np.ones(r.shape)
        indicator[r>R] =0
        return indicator/V
    else:
        if(r>R):
            return 0.
        else:
            return 1/V

def Nk(k,L=3200.):
    """Number of k modes for a given box size. Default is cc box size."""
    kf = 2.*np.pi/L
    vf = (kf)**3
    dk = kf #fftpower uses fundamental mode for dk by default
    Nk = (4*np.pi*k**2 * dk)/vf
    return Nk

def match(rlow,xlow,rhigh,xhigh,mp=10,extrap=1,npts=100,bare=False):
    """For plotting 2pcfs - interpolate small-scale pair counts and large scale FFT grid 2PCF"""
    """Output is r^2 xi """
    rlow,xlow = rlow[rlow>0],xlow[rlow>0] #check if zero because sometimes that happens for nbodykit 2pcf
    m = mp #match_point
    rconc,xconc = np.concatenate([rlow[rlow<m],rhigh[rhigh>=m]]),np.concatenate([xlow[rlow<m],xhigh[rhigh>=m]])
    r = np.logspace(np.log10(rlow.min()),np.log10(rhigh.max()),npts)
    s = ius(rconc,rconc**2 * xconc,ext=extrap)(r)
    if(bare): #return multiplied by r^2 by default
        s = s/r**2
    return r,s

def Delta(z,mdef,cosmo=Planck15):
    if(mdef=='vir'):
        '''Bryan + Norman 1998 fit'''
        xv = 1- cosmo.Omega_m(z)
        return ((18*np.pi**2 - 82*xv -39*xv**2)/(1-xv) * cosmo.rho_crit(z)*1e10)
    elif(mdef=='200m'):
        return 200*cosmo.rho_m(z)*1e10
    elif(mdef=='200c'):
        return 200*cosmo.rho_crit(z)*1e10
    elif(mdef=='Lag'):
        return cosmo.rho_m(z)*1e10
    elif(mdef=='exc'):
        "Approx. Baldauf++ 13 fitting value for z=0"
        return 30*cosmo.rho_m(z)*1e10
    else:
        print("Mass definition not avaliable!")
        raise ValueError

def rDelta(M,z,mdef='vir',cosmo=Planck15):
    "Choosing vir since closest to what M_FoF finds with ll=0.2 (White 2000 Table 1)"
    return ((3/(4*np.pi*Delta(z,mdef,cosmo)))*M)**(1/3)

def mDelta(r,z,mdef='vir',cosmo=Planck15):
    return 4/3 *np.pi*Delta(z,mdef,cosmo)*r**3

@np.vectorize
def sigma(M,z,mdef,P=None,kmin=1e-5,kmax=1e2,num_pts=100,cosmo=Planck15):
    '''
    Get sigma from P using trapezoidal rule
    Input:
    M: Mass defining smoothing scale
    z: redshift
    mdef: choice of mass definition for smoothing window
    optional
    P: Power spectrum callable, if none use linear
    kmin: lower integration range
    kmax: upper integration range
    num_pts: points to use in integration
    '''
    growth = (1/(1+z))
    if P is None:
        kk,Pkk = np.loadtxt('/Users/jsull/Cosmology_Codes/flowpm/flowpm/data/Planck15_a1p00.txt',unpack=True)
        def P(k):
            return np.interp(k,kk,Pkk)

    k = np.logspace(np.log10(kmin),np.log10(kmax),num_pts)

    """using EdS growth"""
    def I(k):
        I = k**2 * P(k) * np.abs(W_TH(k,rDelta(M,z,mdef)))**2
        return I
    Ig = growth*np.sqrt((1/(2*np.pi**2))* np.trapz(I(k),x=k))

    return Ig



#standard dm HZPT
from gzpt import hzpt
from gzpt.bb import *
import numpy as np
from scipy.special import hyp0f1,gamma
from numpy import inf


class Correlator(hzpt):
    def __init__(self,params,hzpt):
        '''
        Parameters
        ----------
        params : list (float)
            List of current hzpt parameter values.
        hzpt : hzpt
            Hzpt base class that holds the linear, zeldovich, and z information.
        '''
        self.params = params
        self.nmax=(len(self.params)-1)//2
        assert self.nmax<=3
        #probably a better way to do this...
        self.hzpt = hzpt

    def Power(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            Power spectrum in k (h/Mpc)
        '''
        def Pk(k):
            if(wantGrad):
                bb,bbgrad = PBB(k,self.params,nmax=self.nmax,wantGrad=True)
                return self.hzpt.P_zel(k) + bb, bbgrad
            else: return self.hzpt.P_zel(k) + PBB(k,self.params,nmax=self.nmax)
        return Pk

    def Xi(self,wantGrad=False):
        '''
        Parameters
        ----------
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns
        ---------
        callable
            2pcf function of r (Mpc/h)
        '''
        def xi(r):
            if(wantGrad):
                bb,bbgrad = XiBB(r,self.params,nmax=self.nmax,wantGrad=True)
                return self.hzpt.Xi_zel(r) + bb, bbgrad
            else: return self.hzpt.Xi_zel(r) + XiBB(r,self.params,nmax=self.nmax)
        return xi

    def wp(self,r,pi_bins=np.linspace(0,100,10,endpoint=False),wantGrad=False):
        "FIXME: Lazy copy from AutoCorrelator - should make this function accessible by both. - general correlator class..."
        """Projected correlation function.
        Parameters:
        -----------
        r: array (float)
            3D abscissa values for xi
        pi_bins: array (float)
            Array must be of size that divides evenly into r - projection window, tophat for now
        wantGrad : boolean,optional
            Whether to return the function value, or the tuple (val,grad).
        Returns:
        ----------
        (array (float), array (float), [array (float)])
            projected radius rp, projected correlation function wp, gradient if wanted
        """
        #Almost as in corrfunc code of Sinha & Garrison
        dpi = pi_bins[1]-pi_bins[0]
        if(wantGrad):
            xi,grad_xi = self.Xi(wantGrad=wantGrad)(r)
        else:
            xi = self.Xi(wantGrad=wantGrad)(r)
        wp = np.zeros(int(len(xi)/len(pi_bins)))
        if(wantGrad): wp_grad=np.zeros((len(wp),len(self.params)))

        #sum over los in each bin
        for i in range(len(wp)-1):
            wp[i] = 2* dpi * np.sum(xi[i*len(pi_bins):(i+1)*len(pi_bins)])
            if(wantGrad): wp_grad[i] = 2* dpi * np.sum(grad_xi[i*len(pi_bins):(i+1)*len(pi_bins)])
        rp = r[::len(pi_bins)]
        #TODO: Make this branching nicer
        if(wantGrad):
            return rp,wp,wp_grad
        else:
            return rp,wp



#tie stuff together around the hzpt object
import numpy as np
import warnings
from gzpt.zel import *
from scipy.interpolate import InterpolatedUnivariateSpline as ius
import os.path as path

class hzpt:
    """
    Parent object for the extended HZPT model for Power spectrum and 2PCF for dark matter and tracers.
    Each hzpt instance corresponds to one cosmology and one (effective) redshift.

    Parameters
    ----------
    cosmo: (cosmology object)
    z: redshift, float

    Methods
    ----------
    __init__

    Example
    --------
    >>> k,plin = np.loadtxt("my_linear_power_file.txt",unpack=True)
    >>> z = 0.5
    >>> model = gzpt.hzpt(k,plin)
    >>> mm = matter.Correlator(params, model)
    >>> gg = tracers.AutoCorrelator(params,model)
    >>> gm = tracers.CrossCorrelator(params,model)

    #3D stats
    r = np.logspace(-1,3)
    ximm = mm.Xi()(r)
    k = np.logspace(-3,1)
    Pgg = gg.Power()(k)

    """
    #For now will only support single z, but can come back to this later
    def __init__(self,klin,plin):
        self.plin = loginterp(klin,plin) #interpolator

        #compute ZA
        self.cleft = CLEFT(klin,plin)
        self.cleft.make_ptable()
        kza,pza = self.cleft.pktable.T #evaluated at klin
        self.P_zel = loginterp(kza,pza) #callable
        rxi = np.logspace(-1,3,4000) #matching up with the SBT in zel
        xiza = self.cleft.compute_xi_real(rxi)
        self.Xi_zel = loginterp(rxi,xiza) #callable



import numpy as np
"The broadband beyond Zeldovich function"

def PBB(k,params,nmax,wantGrad=False):
    '''Pade Power
    Parameters:
    ----------
    k: array
        wavenumbers
    params: list
        hzpt parameters, floats
    nmax: int, optional
        order of Pade expansion
    wantGrad: boolean, optional
        whether to return function or (function,gradient)
    Returns:
    ---------
    array (of size k)
        broadband function value
    tuple of arrays, optional
        broadband function value and gradient wrt hzpt parameters.

    '''

    if(nmax==0):
        A0,R = params
        R1h,R1sq,R2h = None, None, None
    elif(nmax==1):
        A0,R,R1h = params
        R1sq,R2h = None, None
    elif(nmax==2):
        A0,R,R1h,R1sq,R12 = params
        R2h = R1h/(np.sqrt(2)*R12) #R12 is R1h/R2h/sqrt(2) and is forced to be > 1
        if(R12<=.99) : raise ValueError("R12 must be greater than 1 for physical values.")

    def F_comp(k,R):
        """For now only Lorentzian compensation"""
        Fk = 1-1/(1+(k*R)**2)
        return Fk

    def pade(k,R1h,R1sq,R2h):
        if(nmax==0):
            return 1.
        if(nmax==1):
            return 1/(1+(k*R1h)**2)
        elif(nmax==2):
            return (1+ R1sq * k**2) /(1+(k*R1h)**2 + (k*R2h)**4)


    def gradient(k,A0,R,R1h,R1sq,R2h):
        A_grad = F_comp(k,R) * pade(k,R1h,R1sq,R2h)
        if(nmax==0):
            R_grad = (2*A0* k**2)/(R + k**2 * R**3) * F_comp(k,R)
            return np.array([A_grad,R_grad])
        if(nmax==1):
            R_grad = (2*A0)/((1 + k**2 * R1h**2)* k**2 * R**3) * F_comp(k,R)**2
            R1h_grad = -(2*A0* k**2 *R1h)/(1 + k**2 * R1h**2)**2 * F_comp(k,R)
            return np.array([A_grad,R_grad,R1h_grad])
        elif(nmax==2):
            k2,k4 = k**2,k**4
            R1h2,R2h4 = R1h**2,R2h**4
            R_grad = 2*A0*(1+ k2 * R1sq)/((1 + k2 * R1h2 + k4 * R2h4)* k2 * R**3) * F_comp(k,R)**2
            R1h_num = -2*A0* (k2 + 2*(R2h4/R1h2)*k4) * R1h * (1 + k2 * R1sq ) * F_comp(k,R)
            R1h_denom = (1 + k2 * R1h2 + k4 * R2h4)**2
            R1h_grad = R1h_num/R1h_denom
            R1sq_grad = (A0* k2)/(1 + k2 * R1h2 +k4 * R2h4) * F_comp(k,R)
            R2h_grad = -4*A0* k4 * (1+k2 * R1sq) *R2h**3 /(1 + k2 * R1h2 + k4 *R2h4)**2 * F_comp(k,R)

            #convert R2h_grad to R12_grad - R12 = \frac{R1h}{R2h} => \frac{}\partial(PBB(...,R2h))}{\partial{R12}} = grad_R2h * \frac{\partial{R2h}}{\partial{R12}}
            #this last factor is chain_fac and is just \frac{-1*R1h}{sqrt{2} R12^2} = \frac{-sqr{2}*R2h^2}{R1h}
            chain_fac = -np.sqrt(2)*R2h**2 /R1h
            R12_grad = R2h_grad*chain_fac
            #d(pbb)/d(R1h) = A0 F(k) d(bb)/dR1h, d(bb)/dR1h = (1+R1sq k^2) d((1 + k^2 R1h^2 + k^4 R2h^4)^(-1))/d(R1h)

            return np.array([A_grad,R_grad,R1h_grad,R1sq_grad,R12_grad])


    Pbb = A0 * F_comp(k,R) * pade(k,R1h,R1sq,R2h)

    if(wantGrad==True):
        return Pbb, gradient(k,A0,R,R1h,R1sq,R2h).T
    else:
        return Pbb


def XiBB(r,params,nmax,wantGrad=False):
    '''Pade Correlation Function
    Parameters:
    ----------
    r: array
        scale
    params: list
        hzpt parameters, floats
    nmax: int, optional
        order of Pade expansion
    wantGrad: boolean, optional
        whether to return function or (function,grad)
    Returns:
    ---------
    array (of size k)
        broadband function value
    tuple of arrays, optional
        broadband function value and gradient wrt hzpt parameters.

    '''

    if(nmax==0):
        A0,R = params
        R1h,R1sq,R2h=None,None,None
    elif(nmax==1):
        A0,R,R1h = params
        R1sq,R2h = None, None
    elif(nmax==2):
        A0,R,R1h,R1sq,R12 = params
        R2h = R1h/(np.sqrt(2)*R12) #R12 is R1h/R2h/sqrt(2) and is forced to be > 1
        if(R12<=.99) : raise ValueError("R12 must be greater than 1. for physical values.")
    else:
        raise NotImplementedError("nmax>2 not supported for xi")

    def F2_comp(r,R):
        '''Exponential `compensation' - analytic fourier transform of F_comp'''
        F2 = np.exp(-r/R) / (4*np.pi*r*R**2)
        return F2


    def pade2(r,R,R1h,R1sq,R2h):
        if(nmax==0):
            return 1
        elif(nmax==1):
            return (1 - (R/R1h)**2 * np.exp(-(R-R1h)*r/(R*R1h)))/(1 - (R1h/R)**2)
        elif(nmax==2):
            '''Hand et al. 2017 eqns. (B.8-B.11) '''
            pre = 1/(1 -(R1h/R)**2 +(R2h/R)**4)
            S = np.sqrt(R1h**4 - 4*R2h**4)
            A = (1/(2*R2h**4 *S))*(R**2 *(-2*R2h**4 +R1sq *(R1h**2 -S)) + R2h**4 *(R1h**2 -S) + R1sq *(-R1h**4 +2*R2h**4 + R1h**2 *S))
            B = -(1/(2*R2h**4 *S))*(R2h**4 *(R1h**2 + S) - R1sq *(R1h**4 - 2*R2h**4 +R1h**2 *S) + R**2 *(-2*R2h**4 + R1sq *(R1h**2 + S)))

            first = 1- R1sq*(1/R)**2
            second = A*np.exp(r*(1/R - (1/R2h**2)*np.sqrt((R1h**2 - S)/2)))
            third = B*np.exp(r*(1/R - (1/R2h**2)*np.sqrt((R1h**2 + S)/2)))

            return pre*(first + second + third)

    def gradient(r,A0,R,R1h,R1sq,R2h):
        A_grad = - F2_comp(r,R) * pade2(r,R,R1h,R1sq,R2h)
        F_comp_R_grad = -A0 * (r - 2.*R)/R**2 *F2_comp(r,R)

        if(nmax==0):
            R_grad = F_comp_grad
            return np.array([A_grad,R_grad])
        if(nmax==1):
            R2,R1h2,R1h4 = R**2,R1h**2,R1h**4
            R_num = -A0*np.exp(-r/R)*(2.*R**3 *(-1. + np.exp(r*(1/R - 1/R1h)))
                                        +r*(R2 - R1h2)
                                       )
            R_denom = 4.*np.pi*r*(R**3 - R*R1h2)**2
            R_grad = R_num/R_denom
            R1h_num = A0*np.exp(-r/R)*(-2. * R1h**5
                                        + np.exp(r*(1/R - 1/R1h))*R2 *(-2.*R2 *R1h
                                                                       +4.*R1h**3
                                                                       +r*(R2 - R1h2)
                                                                       )
                                        )
            R1h_denom = 4.*np.pi*r*R1h4 * (R2 - R1h2)**2
            R1h_grad = R1h_num/R1h_denom
            return np.array([A_grad,R_grad,R1h_grad])
        elif(nmax==2):
            "Chain rule for this very long expression"
            """TODO:- should make it so don't need to run this twice - probably refactor the gradient into the actual function, turn into f+j"""
            #setup
            R2,R3,R4,R5,R6 = R**2, R**3, R**4, R**5,R**6
            R1h2,R1h3,R1h4,R1h5,R1h6,R1h7,R1h8 = R1h**2, R1h**3, R1h**4, R1h**5, R1h**6, R1h**7, R1h**8
            R2h2,R2h3,R2h4,R2h5,R2h8 = R2h**2, R2h**3, R2h**4, R2h**5, R2h**8

            pre = 1/(1 -(R1h2/R2) +(R2h4/R4))
            S = np.sqrt(R1h4 - 4*R2h4)
            A = (1/(2*R2h4 *S))*(R2 *(-2*R2h4 +R1sq *(R1h2 -S)) + R2h4 *(R1h2 -S) + R1sq *(-R1h4 +2*R2h4 + R1h2 *S))
            B = -(1/(2*R2h4 *S))*(R2h4 *(R1h**2 + S) - R1sq *(R1h4 - 2*R2h4 +R1h2 *S) + R2 *(-2*R2h4 + R1sq *(R1h2 + S)))

            T1 = 1- R1sq/R2
            EA = np.exp(r*(1/R - (1/R2h2)*np.sqrt((R1h2 - S)/2)))
            EB = np.exp(r*(1/R - (1/R2h2)*np.sqrt((R1h2 + S)/2)))
            T2 = A*EA
            T3 = B*EB

            #R
            pre_R_grad = (-2.*R5*R1h2 + 4*R3*R2h4)/(R4 - R2*R1h2 + R2h4)**2
            T1_R_grad = 2*R1sq/R3
            A_R_grad = R*(R1h2*R1sq -2.*R2h4 -R1sq*S)/(R2h4*S)
            EA_R_grad = -(r/R2)*EA
            B_R_grad = -R*(-2.*R2h4 + R1sq*(R1h2 + S))/(R2h4*S)
            EB_R_grad = -(r/R2)*EB
            T2_R_grad = A_R_grad*EA + A*EA_R_grad
            T3_R_grad = B_R_grad*EB + B*EB_R_grad
            BB_R_grad = pre_R_grad*(T1+T2+T3) + pre*(T1_R_grad + T2_R_grad + T3_R_grad)

            R_grad = F_comp_R_grad*pade2(r,R,R1h,R1sq,R2h) + (- A0 * F2_comp(r,R))*BB_R_grad

            #R1h
            pre_R1h_grad = (2.*R6*R1h2-4.*R4*R2h4)/(R1h*(R4 - R2*R1h2 + R2h4)**2)
            S_R1h_grad = 2.*S/R1h
            A_R1h_grad = (R1sq*(R1h4 - 2.*R2h4 - R1h2*S)
                          + 2.*R2*(-R1h2*R1sq + R2h4 + R1sq*S))/(R1h*R2h4*S)
            EA_R1h_grad = ((r*np.sqrt((R1h2-S)/2.))/(R1h*R2h2)) * EA
            B_R1h_grad = (-R1sq*(R1h4 - 2.*R2h4 + R1h2*S)
                           + 2*R2*(-R2h4 + R1sq*(R1h2 + S)))/(R1h*R2h4*S)
            EB_R1h_grad = ((r*np.sqrt((R1h2+S)/2.))/(R1h*R2h2)) * EB
            T2_R1h_grad = A_R1h_grad*EA + A*EA_R1h_grad
            T3_R1h_grad = B_R1h_grad*EB + B*EB_R1h_grad
            BB_R1h_grad = pre_R1h_grad*(T1+T2+T3) + pre*(T2_R1h_grad + T3_R1h_grad)

            R1h_grad = (-A0*F2_comp(r,R))*BB_R1h_grad

            #R1sq
            T1_R1sq_grad = -1./R2
            A_R1sq_grad = (2.*R2h4 + (R-R1h)*(R+R1h)*(R1h2-S))/(2*R2h4*S)
            B_R1sq_grad = -(2.*R2h4 + (R-R1h)*(R+R1h)*(R1h2+S))/(2*R2h4*S)
            T2_R1sq_grad = A_R1sq_grad*EA
            T3_R1sq_grad = B_R1sq_grad*EB
            BB_R1sq_grad = pre*(T1_R1sq_grad + T2_R1sq_grad + T3_R1sq_grad)

            R1sq_grad = (-A0*F2_comp(r,R))*BB_R1sq_grad

            #R2h
            pre_R2h_grad = -(4.*R4*R2h3)/(R4 - R2*R1h2 + R2h4)**2
            S_R2h_grad = -8.*R2h3/S
            A_R2h_grad = -2.*(-R1h8*R1sq + 6.*R1h4*R1sq*R2h4 - 4.*R1sq*R2h8
                              + R1h6*R1sq*S - 2.*R1h2*(R2h8 + 2.*R1sq*R2h4*S)
                              + R2*(R1h6*R1sq - 6.*R1h2*R1sq*R2h4 - R1h4*R1sq*S
                                    + 4.*(R2h8 + R1sq*R2h4*S)
                                   )
                             )/(R2h5*(S**3))
            EA_R2h_grad = r*(-R1h4 + 2.*R2h4 + R1h2*S)/(R2h3*S*np.sqrt((R1h2-S)/2.)) * EA
            B_R2h_grad = -2.*(R1h8*R1sq - 6.*R1h4*R1sq*R2h4 + 4.*R1sq*R2h8
                              + R1h6*R1sq*S - R2*(R1h6*R1sq - 6.*R1h2*R1sq*R2h4
                                                  + 4.*R2h8 + R1h4*R1sq*S -4.*R1sq*R2h4*S
                                                 )
                              + 2.*R1h2*(R2h8 - 2.*R1sq*R2h4*S)
                             )/(R2h5*(S**3))
            EB_R2h_grad = r*(R1h4 - 2.*R2h4 + R1h2*S)/(R2h3*S*np.sqrt((R1h2+S)/2.)) * EB
            T2_R2h_grad = A_R2h_grad*EA + A*EA_R2h_grad
            T3_R2h_grad = B_R2h_grad*EB + B*EB_R2h_grad
            BB_R2h_grad = pre_R2h_grad*(T1+T2+T3) + pre*(T2_R2h_grad + T3_R2h_grad)

            R2h_grad = (-A0*F2_comp(r,R))*BB_R2h_grad

            #see note for description of chain factor in PBB
            chain_fac = -np.sqrt(2)*R2h**2 /R1h
            R12_grad = R2h_grad*chain_fac

            return np.array([A_grad,R_grad,R1h_grad,R1sq_grad,R12_grad])

    Xibb = -A0 * F2_comp(r,R) * pade2(r,R,R1h,R1sq,R2h)
    if(wantGrad==True):
        return Xibb, gradient(r,A0,R,R1h,R1sq,R2h).T
    else:
        return Xibb



from setuptools import setup
from io import open

# read the contents of the README file
#with open('README.md', "r") as f:
#    long_description = f.read()

def find_version(path):
    with open(path, 'r') as fp:
        file = fp.read()
    import re
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                            file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Version not found")

setup(name='gzpt',
      version = find_version("gzpt/__init__.py"),
      description='Hybrid Analytic Model for Matter and Tracer Two-point Correlators',
     # long_description="See README.md",#long_description,
      url='https://github.com/jmsull/gzpt',
      author='James M Sullivan',
      author_email='jmsullivan@berkeley.edu',
      license='MIT',
      packages=['gzpt'],
      install_requires=['wheel', 'numpy>=1.16.5', 'scipy','pyfftw'],
      tests_require=['numpy>=1.16.5','scipy','pyfftw'],
      extras_require={
        'testing':  ["numpy"],
        },
      python_requires='>=3',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Physics'
        ],
      keywords='cosmology')
